#include <stdint.h>
int main() {
  int value = 5;

  value = 5; // break here
  value = 5;
  value = 5;
  value = 5;
  value = 5;
  value = 5;
  value = 10;
  value = 10;
  value = 10;
  value = 10;
  value = 5;
  value = 7;
  value = 9;

  return value;
}
