;;; Directory Local Variables            -*- no-byte-compile: t -*-
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((c-basic-offset . 4)))
 (swift-mode . ((swift-mode:basic-offset . 4))))
