#include <complex.h>

int main() {
  int complex complex_int = -1 + -2 * I;
  long complex complex_long = -1 + -2 * I;
  long long complex complex_long_long = -1 + -2 * I;

  unsigned complex complex_unsigned = 1 + 2 * I;
  unsigned long complex complex_unsigned_long = 1 + 2 * I;
  unsigned long long complex complex_unsigned_long_long = 1 + 2 * I;
  return 0; // break here
}
