
// Format is Name/required. If 'required' is true then loading will fail if the symbol is missing,
// otherwise loading will continue and the function pointer will be null.

INDEXSTORE_FUNCTION(creation_options_add_prefix_mapping, false)
INDEXSTORE_FUNCTION(creation_options_create, false)
INDEXSTORE_FUNCTION(creation_options_dispose, false)
INDEXSTORE_FUNCTION(error_get_description, true)
INDEXSTORE_FUNCTION(error_dispose, true)
INDEXSTORE_FUNCTION(format_version, true)
INDEXSTORE_FUNCTION(version, false)
INDEXSTORE_FUNCTION(store_create, true)
INDEXSTORE_FUNCTION(store_create_with_options, false)
INDEXSTORE_FUNCTION(store_dispose, true)
#if INDEXSTORE_HAS_BLOCKS
INDEXSTORE_FUNCTION(store_units_apply, true)
#endif
INDEXSTORE_FUNCTION(store_units_apply_f, false)
INDEXSTORE_FUNCTION(unit_event_notification_get_events_count, true)
INDEXSTORE_FUNCTION(unit_event_notification_get_event, true)
INDEXSTORE_FUNCTION(unit_event_notification_is_initial, true)
INDEXSTORE_FUNCTION(unit_event_get_kind, true)
INDEXSTORE_FUNCTION(unit_event_get_unit_name, true)
#if INDEXSTORE_HAS_BLOCKS
INDEXSTORE_FUNCTION(store_set_unit_event_handler, true)
#endif
INDEXSTORE_FUNCTION(store_set_unit_event_handler_f, false)
INDEXSTORE_FUNCTION(store_start_unit_event_listening, true)
INDEXSTORE_FUNCTION(store_stop_unit_event_listening, true)
INDEXSTORE_FUNCTION(store_discard_unit, true)
INDEXSTORE_FUNCTION(store_discard_record, true)
INDEXSTORE_FUNCTION(store_purge_stale_data, true)
INDEXSTORE_FUNCTION(store_get_unit_name_from_output_path, true)
INDEXSTORE_FUNCTION(store_get_unit_modification_time, true)
INDEXSTORE_FUNCTION(symbol_get_language, true)
INDEXSTORE_FUNCTION(symbol_get_kind, true)
INDEXSTORE_FUNCTION(symbol_get_subkind, true)
INDEXSTORE_FUNCTION(symbol_get_properties, true)
INDEXSTORE_FUNCTION(symbol_get_roles, true)
INDEXSTORE_FUNCTION(symbol_get_related_roles, true)
INDEXSTORE_FUNCTION(symbol_get_name, true)
INDEXSTORE_FUNCTION(symbol_get_usr, true)
INDEXSTORE_FUNCTION(symbol_get_codegen_name, true)
INDEXSTORE_FUNCTION(symbol_relation_get_roles, true)
INDEXSTORE_FUNCTION(symbol_relation_get_symbol, true)
INDEXSTORE_FUNCTION(occurrence_get_symbol, true)
#if INDEXSTORE_HAS_BLOCKS
INDEXSTORE_FUNCTION(occurrence_relations_apply, true)
#endif
INDEXSTORE_FUNCTION(occurrence_relations_apply_f, false)
INDEXSTORE_FUNCTION(occurrence_get_roles, true)
INDEXSTORE_FUNCTION(occurrence_get_line_col, true)
INDEXSTORE_FUNCTION(record_reader_create, true)
INDEXSTORE_FUNCTION(record_reader_dispose, true)
#if INDEXSTORE_HAS_BLOCKS
INDEXSTORE_FUNCTION(record_reader_search_symbols, true)
INDEXSTORE_FUNCTION(record_reader_symbols_apply, true)
INDEXSTORE_FUNCTION(record_reader_occurrences_apply, true)
INDEXSTORE_FUNCTION(record_reader_occurrences_in_line_range_apply, true)
INDEXSTORE_FUNCTION(record_reader_occurrences_of_symbols_apply, true)
#endif
INDEXSTORE_FUNCTION(record_reader_search_symbols_f, false)
INDEXSTORE_FUNCTION(record_reader_symbols_apply_f, false)
INDEXSTORE_FUNCTION(record_reader_occurrences_apply_f, false)
INDEXSTORE_FUNCTION(record_reader_occurrences_in_line_range_apply_f, false)
INDEXSTORE_FUNCTION(record_reader_occurrences_of_symbols_apply_f, false)
INDEXSTORE_FUNCTION(unit_reader_create, true)
INDEXSTORE_FUNCTION(unit_reader_dispose, true)
INDEXSTORE_FUNCTION(unit_reader_get_provider_identifier, true)
INDEXSTORE_FUNCTION(unit_reader_get_provider_version, true)
INDEXSTORE_FUNCTION(unit_reader_get_modification_time, true)
INDEXSTORE_FUNCTION(unit_reader_is_system_unit, true)
INDEXSTORE_FUNCTION(unit_reader_is_module_unit, true)
INDEXSTORE_FUNCTION(unit_reader_is_debug_compilation, true)
INDEXSTORE_FUNCTION(unit_reader_has_main_file, true)
INDEXSTORE_FUNCTION(unit_reader_get_main_file, true)
INDEXSTORE_FUNCTION(unit_reader_get_module_name, true)
INDEXSTORE_FUNCTION(unit_reader_get_working_dir, true)
INDEXSTORE_FUNCTION(unit_reader_get_output_file, true)
INDEXSTORE_FUNCTION(unit_reader_get_sysroot_path, true)
INDEXSTORE_FUNCTION(unit_reader_get_target, true)
INDEXSTORE_FUNCTION(unit_dependency_get_kind, true)
INDEXSTORE_FUNCTION(unit_dependency_is_system, true)
INDEXSTORE_FUNCTION(unit_dependency_get_filepath, true)
INDEXSTORE_FUNCTION(unit_dependency_get_modulename, true)
INDEXSTORE_FUNCTION(unit_dependency_get_name, true)
INDEXSTORE_FUNCTION(unit_include_get_source_path, true)
INDEXSTORE_FUNCTION(unit_include_get_target_path, true)
INDEXSTORE_FUNCTION(unit_include_get_source_line, true)
#if INDEXSTORE_HAS_BLOCKS
INDEXSTORE_FUNCTION(unit_reader_dependencies_apply, true)
INDEXSTORE_FUNCTION(unit_reader_includes_apply, true)
#endif
INDEXSTORE_FUNCTION(unit_reader_dependencies_apply_f, false)
INDEXSTORE_FUNCTION(unit_reader_includes_apply_f, false)

#undef INDEXSTORE_FUNCTION
