//===----------------------------------------------------------------------===//
//
// This source file is part of the Swift open source project
//
// Copyright (c) 2014-2023 Apple Inc. and the Swift project authors
// Licensed under Apache License v2.0 with Runtime Library Exception
//
// See http://swift.org/LICENSE.txt for license information
// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
//
//===----------------------------------------------------------------------===//

import enum PackageModel.ProductFilter
import struct PackageModel.PackageReference

public struct DependencyResolverBinding {
    public let package: PackageReference
    public let boundVersion: BoundVersion
    public let products: ProductFilter
}
