//===----------------------------------------------------------------------===//
//
// This source file is part of the Swift.org open source project
//
// Copyright (c) 2014 - 2019 Apple Inc. and the Swift project authors
// Licensed under Apache License v2.0 with Runtime Library Exception
//
// See https://swift.org/LICENSE.txt for license information
// See https://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
//
//===----------------------------------------------------------------------===//

// This file is automatically generated with generate-swift-format. Do not edit!

/// By default, the `Rule.ruleName` should be the name of the implementing rule type.
@_spi(Testing)
public let ruleNameCache: [ObjectIdentifier: String] = [
  ObjectIdentifier(AllPublicDeclarationsHaveDocumentation.self): "AllPublicDeclarationsHaveDocumentation",
  ObjectIdentifier(AlwaysUseLiteralForEmptyCollectionInit.self): "AlwaysUseLiteralForEmptyCollectionInit",
  ObjectIdentifier(AlwaysUseLowerCamelCase.self): "AlwaysUseLowerCamelCase",
  ObjectIdentifier(AmbiguousTrailingClosureOverload.self): "AmbiguousTrailingClosureOverload",
  ObjectIdentifier(BeginDocumentationCommentWithOneLineSummary.self): "BeginDocumentationCommentWithOneLineSummary",
  ObjectIdentifier(DoNotUseSemicolons.self): "DoNotUseSemicolons",
  ObjectIdentifier(DontRepeatTypeInStaticProperties.self): "DontRepeatTypeInStaticProperties",
  ObjectIdentifier(FileScopedDeclarationPrivacy.self): "FileScopedDeclarationPrivacy",
  ObjectIdentifier(FullyIndirectEnum.self): "FullyIndirectEnum",
  ObjectIdentifier(GroupNumericLiterals.self): "GroupNumericLiterals",
  ObjectIdentifier(IdentifiersMustBeASCII.self): "IdentifiersMustBeASCII",
  ObjectIdentifier(NeverForceUnwrap.self): "NeverForceUnwrap",
  ObjectIdentifier(NeverUseForceTry.self): "NeverUseForceTry",
  ObjectIdentifier(NeverUseImplicitlyUnwrappedOptionals.self): "NeverUseImplicitlyUnwrappedOptionals",
  ObjectIdentifier(NoAccessLevelOnExtensionDeclaration.self): "NoAccessLevelOnExtensionDeclaration",
  ObjectIdentifier(NoAssignmentInExpressions.self): "NoAssignmentInExpressions",
  ObjectIdentifier(NoBlockComments.self): "NoBlockComments",
  ObjectIdentifier(NoCasesWithOnlyFallthrough.self): "NoCasesWithOnlyFallthrough",
  ObjectIdentifier(NoEmptyTrailingClosureParentheses.self): "NoEmptyTrailingClosureParentheses",
  ObjectIdentifier(NoLabelsInCasePatterns.self): "NoLabelsInCasePatterns",
  ObjectIdentifier(NoLeadingUnderscores.self): "NoLeadingUnderscores",
  ObjectIdentifier(NoParensAroundConditions.self): "NoParensAroundConditions",
  ObjectIdentifier(NoPlaygroundLiterals.self): "NoPlaygroundLiterals",
  ObjectIdentifier(NoVoidReturnOnFunctionSignature.self): "NoVoidReturnOnFunctionSignature",
  ObjectIdentifier(OmitExplicitReturns.self): "OmitExplicitReturns",
  ObjectIdentifier(OneCasePerLine.self): "OneCasePerLine",
  ObjectIdentifier(OneVariableDeclarationPerLine.self): "OneVariableDeclarationPerLine",
  ObjectIdentifier(OnlyOneTrailingClosureArgument.self): "OnlyOneTrailingClosureArgument",
  ObjectIdentifier(OrderedImports.self): "OrderedImports",
  ObjectIdentifier(ReplaceForEachWithForLoop.self): "ReplaceForEachWithForLoop",
  ObjectIdentifier(ReturnVoidInsteadOfEmptyTuple.self): "ReturnVoidInsteadOfEmptyTuple",
  ObjectIdentifier(TypeNamesShouldBeCapitalized.self): "TypeNamesShouldBeCapitalized",
  ObjectIdentifier(UseEarlyExits.self): "UseEarlyExits",
  ObjectIdentifier(UseExplicitNilCheckInConditions.self): "UseExplicitNilCheckInConditions",
  ObjectIdentifier(UseLetInEveryBoundCaseVariable.self): "UseLetInEveryBoundCaseVariable",
  ObjectIdentifier(UseShorthandTypeNames.self): "UseShorthandTypeNames",
  ObjectIdentifier(UseSingleLinePropertyGetter.self): "UseSingleLinePropertyGetter",
  ObjectIdentifier(UseSynthesizedInitializer.self): "UseSynthesizedInitializer",
  ObjectIdentifier(UseTripleSlashForDocumentationComments.self): "UseTripleSlashForDocumentationComments",
  ObjectIdentifier(UseWhereClausesInForLoops.self): "UseWhereClausesInForLoops",
  ObjectIdentifier(ValidateDocumentationComments.self): "ValidateDocumentationComments",
]
