#include <stdio.h>

int main (int argc, char const *argv[])
{
    struct {
        int a;
        int b;
        int c;
    } MyAggregateDataType;

    printf ("Set break point at this line.\n");
    return 0;
}
