import CModule

func use<T>(_ t: T) {}

let pureSwift = 42
let point = Point(x: 1, y: 2)
let enumerator = yellow

struct SwiftStruct {
  let name = "pure swift"
}

let pureSwiftStruct = SwiftStruct()

struct SwiftStructCMember {
  let point = Point(x: 3, y: 4)
  let sub = FromSubmodule(x: 1, y: 2, z: 3)
  let name = "swift struct c member"
}

let swiftStructCMember = SwiftStructCMember()
let typedef = TPoint(x: 5, y: 6)
let union = DoubleLongUnion(long_val: 42)
let fromSubmodule = FromSubmodule(x: 1, y: 2, z: 3)
let withPointer = WithPointer(ptr: UnsafeMutableRawPointer(bitPattern: 0x0),
                        const_ptr: UnsafeMutableRawPointer(bitPattern: 0x0))

use(pureSwift) // break here
use(point)
use(enumerator)
use(pureSwiftStruct)
use(swiftStructCMember)
use(typedef)
use(union)
use(fromSubmodule)
use(withPointer)
