#include <stdio.h>

int main (int argc, char const *argv[])
{
  int pass_me = argc + 10; // Stop here to get started.
  printf("This is the zeroth stop\n");
  printf("This is the first stop\n");
  printf("This is the second stop\n");
  printf("This is the third stop\n");
  printf("This is the fourth stop\n");
  printf("This is the fifth stop\n");
  printf("This is the sixth stop\n");
  printf("This is the seventh stop\n");
  printf("This is the eighth stop\n");
  printf("This is the nineth stop\n");

  return 0;
}
