extern int absent_weak_int __attribute__((weak_import));

extern int present_weak_int __attribute__((weak_import));

extern int absent_weak_function() __attribute__((weak_import));

extern int present_weak_function() __attribute__((weak_import));

